/**
* team_server.c - P2PWNC Team Server module
*
* Copyright (C) 2004-2005 Pantelis Frangoudis, Elias Efstathiou
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


//DONE
//kati prepei na kanw gia permanent storage
//p.x. otan ksekina na diavazei receipts apo arxeio
//ka8e ligo kai ligaki na kanei backup ta nea receipts
//'h ka8e fora pou kaleitai h insert_receipt na th grafei kai se arxeio (vlepe client_rdb)
//'h alliws na th grafei se xwristo arxeio (ka8e rcpt sto diko ths)

#include "io.h"
#include "crypt.h"
#include "conffile.h"
#include "threads.h"
#include "rdb_engine.h"

#define RDB_FILE "../data/receipts.db"
#define RDB_FREE_LIST "../data/receipts.fl"

//configuration
unsigned char str_pub_key_path[80];
unsigned char str_conf_file_path[80];
unsigned char rdb_file[80];
unsigned char rdb_free_list[80];

int server_port;
int max_threads = 50;

#define DEF_CONF_PATH "../conf/p2pwnc_team_server.conf"
#define DEF_CONF_FNAME "p2pwnc_team_server.conf"

typedef struct rbtree_range_query_func_data_t {
	int sockfd;
} RQDATA;

//for each receipt, this func will be called...
//each receipt will be sent to the client in an RCPT message
void range_query_callback(RBTNODE *node, void *data) {
	unsigned char *text = NULL; //text message to be sent

	text = receipt_2_text( ((RBTREE_RCPT*)node->data)->rcpt );
	if (!text) {
		//error constructing rcpt message
		return;
	}

	//write message to socket
	if (!write( ((RQDATA*)data)->sockfd, text, strlen(text))) {
		fprintf(stderr, "range_query: Error writing to socket\n");
	}

	//free message text
	SAFE_FREE(text);
}

PWNC_PUBKEY *server_pub;

struct sockaddr_in server_addr;
int sockfd;

void term_handler(int signum) {
	//final call: clean up openssl structures
	int i = 0;
	for (i = 0; i<max_threads; i++) {
		cancel_thread(i);
	}
	cleanup_ossl_threads();
	thread_pool_cleanup();
	SAFE_FREE_KEY(server_pub);
	exit(0);
}

void client_handler(THREAD_INFO *ti) {
	unsigned char *text = NULL; //text message
	unsigned char *text_data = NULL; //UPDT text message data
	int mtype = 0; //message type
	int mlen = 0;//message length
	int errval = 0; //used to check for errors
	PWNC_RCPT *rcpt = NULL; //in case of an RCPT message (--> store receipt)
	unsigned char *rinfo = NULL; //serialized receipt info
	int info_len = 0; //cert info (serialized) length
	int verified = 0;

	unsigned long timestamp = 0;

	RQDATA *rqd = NULL; //range query data (==> client socket addr)

	struct timeval delay; //delay
	int select_status = 0; //used in "select" calls
	fd_set fs;

	if (block_signals()) {
		fprintf(stderr, "Warning: Could not set signal mask\n");
	}
	
	//get thread info
	int sockfd = ti->client_socket;
	int tindex = ti->thread_index;
	SAFE_FREE(ti);

	//receive message
	text = read_message(sockfd, &mtype);

#ifdef TS_JUDGE
	//the team server can answer QUER messages
	PWNC_PUBKEY *provider = NULL;
	PWNC_PUBKEY *consumer = NULL;
	if (mtype == PWNC_MSG_QUER) {
		ALLOC_AND_CLEAN(provider, sizeof(PWNC_PUBKEY), PWNC_PUBKEY*);
		ALLOC_AND_CLEAN(consumer, sizeof(PWNC_PUBKEY), PWNC_PUBKEY*);
		if (!parse_query_msg(text, provider, consumer) && rdb_judge(consumer, provider)) {
			SAFE_FREE(text);
			//ok, access should be granted..
			text = construct_qrsp(PWNC_ACTION_GRANT);
		}
		else {
			SAFE_FREE(text);
			//access should not be granted or some other error occured
			text = construct_qrsp(PWNC_ACTION_FORBID);
		}
		SAFE_FREE_KEY(provider);
		SAFE_FREE_KEY(consumer);

		//sendQRSP message
		if (!write(sockfd, text, strlen(text))) {
			fprintf(stderr, "team_server: Error writing QRSP to socket\n");
		}

		if (sockfd) {
			close(sockfd);
		}

		SAFE_FREE(text);
		//pthread_exit(NULL);
		exit_client_thread(tindex);
	}
	else
#endif

	if ( (mtype != PWNC_MSG_UPDT) && (mtype != PWNC_MSG_RCPT) ) {
		fprintf(stderr, "Invalid message type\n");
		SAFE_FREE(text);
		exit_client_thread(tindex);
		return;
	}

	//receipt message
	//i. verify receipt
	//ii. store it in rdb
	if (mtype == PWNC_MSG_RCPT) {
		rcpt = text_2_rcpt(text);
		if (!rcpt) {
			fprintf(stderr, "Error parsing receipt\n");
			exit_client_thread(tindex);
			return;
		}
		//verification of receipt signature
		rinfo = serialize_receipt_info(rcpt->ConsumerCert, rcpt->ProviderPubKey, rcpt->timestamp, rcpt->weight, &info_len);
		if (!rinfo) {
			//err
			fprintf(stderr, "Error serializing receipt info\n");
			//end_client_session(sockfd);
			SAFE_FREE_RCPT(rcpt);
			exit_client_thread(tindex);
			return;
		}
		verified = verify_pwnc_signature(rcpt->RcptSig, rcpt->ConsumerCert->UserPubKey, rinfo, info_len, 1);

		if (verified != 1) {
			fprintf(stderr, "Receipt verification failed\n");
		}
		else {
			insert_receipt(rcpt, 1); //TODO: check arg --> maybe the ts doesn't need to know how the sublayer stores receipts (maybe a wrapper func here??)
			//akyro mallon, DONE, an einai compiled me DPERM_STORAGE tote to rdb 8a taktopoihsei ta pragmata, 8a to grapsei sto disko
			//alliws, to (...,1) 8a agnoh8ei
		}

		exit_client_thread(tindex);
		return;
	}

	//parse UPDT message ---> get timestamp
	errval = parse_updt_req(text, &timestamp);
	if (errval == -1) {
		SAFE_FREE(text);
		exit_client_thread(tindex);
		return;
	}
	SAFE_FREE(text);

	//so far so good....

	//construct range query data
	rqd = (RQDATA*)malloc(sizeof(RQDATA));
	memset(rqd, 0, sizeof(RQDATA));
	rqd->sockfd = sockfd;

	//range query (sends receipts to client)
	rdb_range_query( range_query_callback, (void*)rqd, timestamp );
	SAFE_FREE(rqd);

	//send UPDT end message after finishing with all receipts
	//construct message
	text_data = make_updt_end_data();
	if (!text_data) {
		fprintf(stderr, "Error constructing UPDT text data\n");
		exit_client_thread(tindex);
		return;
	}
	text = construct_updt_msg(text_data, 3);
	SAFE_FREE(text_data);
	if (!text) {
		fprintf(stderr, "Error constructing UPDT end\n");
		exit_client_thread(tindex);
		return;
	}

	//send it...
	if (!write(sockfd, text, strlen(text))) {
		fprintf(stderr, "Error writing to socket\n");
	}
	SAFE_FREE(text);
	exit_client_thread(tindex);
	//exit thread
}


int read_keys(const char *pub_key_path) {
	server_pub = read_key_file(pub_key_path);
	if (!server_pub) {
		return 0;
	}
	return 1;
}

int init_server(int portnum) {
	memset(&server_addr, 0, sizeof(server_addr));
	server_addr.sin_family = AF_INET;
	server_addr.sin_addr.s_addr = INADDR_ANY;
	//server_addr.sin_port = htons(PWNC_SERVER_PORT);
	server_addr.sin_port = htons(portnum);

	sockfd = socket (PF_INET, SOCK_STREAM, 0);
	if (sockfd < 0) {
		fprintf(stderr, "socket creation failed\n");
		return 0;
	}

	//Bind a local address to the socket
	if (bind(sockfd, (struct sockaddr *)&server_addr, sizeof (server_addr)) < 0) {
		fprintf(stderr,"bind failed\n");
		return 0;
	}

	//Specify a size of request queue
	if (listen(sockfd, 20) < 0) {
		fprintf(stderr,"listen failed\n");
		return 0;
	}

	//db initialization: initialize rbtree, hash tables, etc
	rdb_init(rdb_file, rdb_free_list);
	rdb_load_data();

	return 1;
}

void print_usage() {
	fprintf(stderr, "\nUsage: team_server ");
	fprintf(stderr, "[-h/--help] ");
	fprintf(stderr, "[-c/--conf file] ");
	fprintf(stderr, "[-pk/--public-key-path file] ");
	fprintf(stderr, "[-sp/--server-port port] ");
	fprintf(stderr, "[-mt/--max-threads number_of_threads] ");
	fprintf(stderr, "[-rdb/--rdb-file path] ");
	fprintf(stderr, "[-rfl/--rdb-free-list path]\n\n");

	fprintf(stderr, "\t-h --help\t\tDisplay this message\n");
	fprintf(stderr, "\t-c --conf\t\tTeam server configuration file\n");
	fprintf(stderr, "\t-pk --public-key-path\tTeam public key file\n");
	fprintf(stderr, "\t-sp --server-port\tTeam server port\n");
	fprintf(stderr, "\t-mt --max-threads\tMaximum number of threads\n");
	fprintf(stderr, "\t-rdb --rdb-file\t\tReceipt database file\n");
	fprintf(stderr, "\t-rfl --rdb-free-list\tReceipt database list of free positions file\n\n");
}

int main(int argc, char **argv) {
	struct sockaddr_in client_addr;
	int sockfd2;
	int alen = 0;
	FILE *conf_file_fp = NULL;

	memset(&str_pub_key_path, 0, 80);
	memset(&str_conf_file_path, 0, 80);
	memset(&rdb_file, 0, 80);
	memset(&rdb_free_list, 0, 80);

	int help = 0;

	CMD_ARG *cmd_line_args = NULL;
	CENTRY *conf_entries = NULL;
	unsigned char *valid_conf_names[] = {
		"public-key-path",
		"server-port",
		"max-threads",
		"rdb-file",
		"rdb-free-list"
		};
	unsigned char *valid_cmd_line_names[] ={
		"conf",
		"public-key-path",
		"server-port",
		"max-threads",
		"rdb-file",
		"rdb-free-list",
		"help"
		};
	unsigned char* cmd_line_switches[] = {"c", "pk", "sp", "mt", "rdb", "rfl", "h"};
	unsigned short conf_var_types[] = {VTYPE_STRING, VTYPE_INT, VTYPE_INT, VTYPE_STRING, VTYPE_STRING};
	unsigned short cmd_line_var_types[] = {VTYPE_STRING, VTYPE_STRING, VTYPE_INT, VTYPE_INT, VTYPE_STRING, VTYPE_STRING, VTYPE_NOVAL};
	void *conf_variables[] = {&str_pub_key_path, &server_port, &max_threads, &rdb_file, &rdb_free_list};
	void *cmd_line_variables[] = {&str_conf_file_path, &str_pub_key_path, &server_port, &max_threads, &rdb_file, &rdb_free_list, &help};
	conf_entries = make_valid_entries(valid_conf_names, conf_variables, conf_var_types, 5);
	cmd_line_args = make_valid_args(valid_cmd_line_names, cmd_line_switches, cmd_line_variables, cmd_line_var_types, 7);

	if (argc < 2) {
		//no command line flags specified, so parse default conf file
		conf_file_fp = fopen(DEF_CONF_PATH, "r");
		if (!conf_file_fp) {
			//first check in /etc/pwnc.conf
			conf_file_fp = fopen(DEF_CONF_FNAME, "r");
			if (!conf_file_fp) {
				//if fopen failed, try to open pwnc.conf in the current path
				//if all else fails ---> exit
				fprintf(stderr, "team_server: Error opening configuration file\n");
				print_usage();
				exit(1);
			}
		}
		//default conffile opened ok, so parse it
		parse_conf_file(conf_entries, 5, conf_file_fp);
	}
	else {
		//parse command line
		//if there is a -c flag, ignore all other flags and check the specified conf file
		parse_cmd_line_args(argc, argv, cmd_line_args, 7);
		if (help) {
			print_usage();
			exit(0);
		}
		if (strlen(str_conf_file_path)) {
			//there was a -c flag, so ignore all other flags and check this conf file
			conf_file_fp = fopen(str_conf_file_path, "r");
			if (!conf_file_fp) {
				fprintf(stderr, "team_server: Warning: Failed to open %s, trying defaults\n", str_conf_file_path);
				//try to open specified conffile - if it fails, try defaults again
				conf_file_fp = fopen(DEF_CONF_PATH, "r");
				if (!conf_file_fp) {
					//trying default conffile path
					conf_file_fp = fopen(DEF_CONF_FNAME, "r");
					if (!conf_file_fp) {
						//trying to open default conf file in current dir
						fprintf(stderr, "team_server: Error opening configuration file\n");
						print_usage();
						exit(1);
					}
				}
			}
			parse_conf_file(conf_entries, 5, conf_file_fp);
		}
		//otherwise: we are using cmd line arguments
	}

	//check mandatory fields
	if (!strlen(str_pub_key_path) || !server_port) {
		fprintf(stderr, "team_server: Some configuration info is missing\n");
		print_usage();
		exit(1);
	}

	//if rdb paths have not been specified, use defaults
	if (!rdb_file) {
		memcpy(rdb_file, RDB_FILE, strlen(RDB_FILE));
	}
	if (!rdb_free_list) {
		memcpy(rdb_free_list, RDB_FREE_LIST, strlen(RDB_FREE_LIST));
	}

	fprintf(stderr, "Team server configuration\n");
	fprintf(stderr, "-------------------------\n");
	fprintf(stderr, "Team public key path: %s\n", str_pub_key_path);
	fprintf(stderr, "Team server port: %d\n", server_port);
	fprintf(stderr, "Thread pool size: %d\n", max_threads);
	fprintf(stderr, "Receipt database file: %s\n", rdb_file);
	fprintf(stderr, "Receipt database free positions file: %s\n\n", rdb_free_list);

	if (!read_keys(str_pub_key_path)) {
		fprintf(stderr, "Error reading team public key\n");
		exit(1);
	}

	//seed prng
	init_prng();

	init_ossl_threads();
	
	//initialize thread pool and thread pool mutex
	init_thread_pool(max_threads);
	pthread_mutex_init(&thread_pool_mutex, NULL);

	//install signal handlers
	signal(SIGINT, term_handler);
	signal(SIGTERM, term_handler);

	if (!init_server(server_port)) {
		SAFE_FREE_KEY(server_pub);
		exit(1);
	}

	alen = sizeof(client_addr);

	while (1) {
		fprintf(stderr, "P2PWNC team server: waiting for connections...\n");

		if (  (sockfd2=accept(sockfd, (struct sockaddr *)&client_addr, &alen)) < 0) {
			fprintf(stderr, "accept failed\n");
			//exit (1);
			continue;
		}
		start_new_thread((void*)&client_handler, sockfd2);
	}
	close(sockfd);
	return 0;
}
